#include <iostream>
#include <cstdlib>
#include <cctype>

#include "zero.hh"
#include "kertoimet.hh"
#include "vakiot.hh"
#include "yksikot.hh"
#include "yksikko.hh"
#include "summa.hh"
#include "tulo.hh"

extern void Dump(elemp res);

Kertoimet         Kert;
Yksikot           Yks;
Yksikkomuunnokset Ymu(Yks);
Vakiot            Vak(Yks);

class Parser
{
public:
	static void delblank(string &s)
	{
		while(s.size() && isspace(s[0]))s.erase(0, 1);
	}
	static merknum fetchnum(string &str) /* const is not valid for static */
	{
		char *ep;
		const char *s = str.c_str();
		const char *const begin = s;
		
		real val = strtold(s, &ep);
		
		/* Got the value, next analyze the precision */
		unsigned precis = 0;
		
		const bool VirallinenTyyli = true;
		
		const char *q = s;
		if(*q == '+' || *q == '-')++q;
		bool pilkku = false;
		unsigned nollia=0;
		for(; q < ep; ++q)
		{
			if(*q == '0') { if(pilkku&&(VirallinenTyyli||precis))++precis;else ++nollia; continue; }
			if(*q >= '1' && *q <= '9')
			{
				if(!pilkku)precis += nollia;
				++precis;
				nollia = 0;
				continue;
			}
			/* Do not analyze the exponent */
			if(*q == 'E' || *q == 'e')break;
			if(*q == '.')
			{
				if(precis)precis += nollia;
				pilkku = true;
				nollia = 0;
			}
		}
		if(VirallinenTyyli || pilkku)
		{
			// Virallisessa 1000:n tarkkuus on 4
			// Joka tapauksessa 0.100:n tarkkuus on 3
			precis += nollia;
		}
		
		str.erase(0, ep-begin);
		delblank(str);
		
		return merknum(val, precis);
	}
	static tulo *ParseTulo(string &s, merknum &kerroin)
	{
		tulo *tulotmp = new tulo;
		
		bool jako = false;
		
		for(;;)
		{
    		delblank(s);
    		if(!s.size())break;
    		
    		unsigned a, b;
    		int k = 0;
    		
    		elemp res;
    		
    		merknum ker = fetchnum(s);
    		if(!ker.valid())ker = 1;
    		
    		a = Kert.Test(s);
    		if(a && (b = Yks.Test(s, a)))
    		{
    			k = Kert.Value(s);
    			/* k:n yksikk: 10^ */
    		YksLabel:
    			char y = Yks.Value(s, a);
    			s.erase(0, a+b);
    			
    			ker.pow10(k);
    			res = new yksikko(y);
    		}
    		else if((a = Vak.Test(s)))
			{
				res = &*Vak.Value(s);
				s.erase(0, a);
				
				ker.maxprecision();
			}
			else /* a=0 */
			{
				b = Yks.Test(s);
				if(b)
				{
					/* k = 0; */
					goto YksLabel;
				}
				/* a=0, b=0 */
				res = NULL;
				
				if(s.size() && s[0] == '(')
				{
					s.erase(0, 1);
					res = ParseSumma(s);
					delblank(s);
					if(s[0] != ')')return NULL;
					s.erase(0, 1);
				}
			}
			
			if(jako)
				kerroin /= ker;
			else
	    		kerroin *= ker;
    		
    		if(!res)break;
    		
    		real pot = 1;
    		delblank(s);
    		if(s.size() && s[0]=='^')
   			{
   				s.erase(0, 1);
   				delblank(s);
   				pot = fetchnum(s).value();
    		}
    		tulotmp->insert(res, jako ? -pot : pot);
    		
    		jako = false;
    		if(s.size() && s[0]=='*')
    		{
    			s.erase(0, 1);
    		}
    		if(s.size() && s[0]=='/')
    		{
    			jako = true;
    			s.erase(0, 1);
    		}
		}
		return tulotmp;
	}
	static elemp ParseSumma(string &s)
	{
		summa *summatmp = new summa;
		bool miinus = false;
		for(;;)
		{
    		delblank(s);
    		if(!s.size())break;
    		
			merknum kerroin = 1;
			tulo *p = ParseTulo(s, kerroin);
			if(!p)break;
			
			if(p->size())
				summatmp->insert(p, miinus ? -kerroin : kerroin);
			else if(miinus)
				summatmp->lisays -= kerroin;
			else
				summatmp->lisays += kerroin;
			
			delblank(s);
			if(s.size() && s[0]=='+')
			{
				miinus = false;
				s.erase(0, 1);
			}
			else if(s.size() && s[0]=='-')
			{
				miinus = true;
				s.erase(0, 1);
			}
			else break;
		}
		return summatmp;
	}
	
	static void DumpMerkNum(const merknum &e)
	{
		char Buf[256];
		sprintf(Buf, "%.*Lf", (int)e.precision(), e.value());
		cout << Buf;
	}
	
	static void DumpResult(elemp result)
	{
	    cout << "\33[1;32m" << result->type << "\33[m";
		switch(result->type)
		{
			case '+':
			{
				summa &s = *(summa *)&*result;
				
				cout << '(';
				
				bool iniplus = false;
				
				if(s.lisays != 0 || s.elems.size() == 0)
				{
					DumpMerkNum(s.lisays);
					iniplus = true;
				}
				
				for(summa::elemap::const_iterator i = s.elems.begin(); i != s.elems.end(); ++i)
				{
					if(i->second >= 0)
					{
						if(iniplus)cout << " + ";
						DumpMerkNum(i->second);
					}
					else
					{
						cout << " - ";
						DumpMerkNum(-i->second);
					}
					DumpResult(i->first);
					iniplus = true;
				}
				
				cout << ')';
				break;
			}
			case '*':
			{
				tulo &s = *(tulo *)&*result;
				
				tulo::elemap::const_iterator i;
				
				bool eka=true;
				
				for(i = s.elems.begin(); i != s.elems.end(); ++i)
				{
					if(i->second < 0)continue;
					DumpResult(i->first);
					if(i->second != 1)cout << '^' << i->second;
					eka = false;
				}
				for(i = s.elems.begin(); i != s.elems.end(); ++i)
				{
					if(i->second > 0)continue;
					if(eka) { cout << '1'; eka = false; }
					cout << '/';
					DumpResult(i->first);
					if(i->second != -1)cout << '^' << -i->second;
				}
				break;
			}
			
			default:
				cout << Yks[result->type];
		}
	}
	
	static void Parse(string s)
	{
		elemp result = ParseSumma(s);
		bool vihre = false;
		if(s.size())
		{
			cerr << "Error, left \"" << s << "\"\n";
			vihre = true;
		}
		if(!result)
		{
			cerr << "No result!\n";
			vihre = true;
		}
		if(vihre)return;
		
		cout << "Before optimization: ";
		Dump(result);
		
		result = result->Expand();
		
		cout << "Result: ";		
		Dump(result);		
		/*
		result = result->Expand();
		
		cout << "Once more: ";
		Dump(result);
		*/
	}
};

static Parser p;
void Dump(elemp res)
{
	p.DumpResult(res);
	cout << endl;
}

int main()
{
	string tmp = "5kgc^2";
	
	p.Parse(tmp);
	
	return 0;
}
